#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 24 16:04:06 2019

@author: farid
"""
import numpy as np
import os
from scipy import stats
import seaborn as sns
import matplotlib.patches as patches
import matplotlib.colors as mcolors
import matplotlib.pylab as plt
from .bayes_inference import BayesInference

# Load the mplstyle
plt.style.use(os.path.join(os.path.split(__file__)[0],
                           '../', 'bayesvalidrox.mplstyle'))


class BayesModelComparison:
    def __init__(self, justifiability=True):
        self.justifiability = justifiability
        self.SamplingMethod = 'rejection'

    # --------------------------------------------------------------------------
    def create_model_comparison(self, modelDict, optsDict):

        if not isinstance(modelDict, dict):
            raise Exception("To run model comparsion, you need to pass a "
                            "dictionary of models.")

        # Extract model names
        self.model_names = [*modelDict]

        # Compute total number of the measurement points
        MetaModel = list(modelDict.items())[0][1]
        self.n_meas = MetaModel.ModelObj.n_obs

        # Generate justifiability data
        justData = self.generate_justData(modelDict)

        # Run create Interface for each model
        bayesDict = {}
        for model in modelDict.keys():
            print("-"*20)
            print("Bayesian inference of {}.\n".format(model))

            BayesOpts = BayesInference(modelDict[model])

            # Set BayesInference options
            for key, value in optsDict.items():
                if key in BayesOpts.__dict__.keys():
                    setattr(BayesOpts, key, value)

            # Pass justifiability data as perturbed data
            BayesOpts.perturbed_data = justData
            BayesOpts.just_analysis = self.justifiability

            bayesDict[model] = BayesOpts.create_inference()
            print("-"*20)

        # Compute model weights
        BME_Dict = dict()
        for modelName, bayesObj in bayesDict.items():
            BME_Dict[modelName] = np.exp(bayesObj.log_BME)

        # TODO: BME correction in BayesInference class
        modelWeights = self.cal_modelWeight(BME_Dict)

        # Plot model weights
        if self.justifiability:
            model_names = self.model_names
            model_names.insert(0, 'Observation')

            # Split the model weights and save in a dict
            list_ModelWeights = np.split(
                modelWeights, modelWeights.shape[1]/self.n_meas, axis=1)
            ModelWeights_dict = {key: weights for key, weights in
                                 zip(model_names, list_ModelWeights)}

            self.plot_JustAnalysis(ModelWeights_dict)
        else:
            self.boxPlot_ModelWeights(modelWeights, 'model_weights')

            # Plot kde plot for model weights
            self.kdePlot_BayesFactor(BME_Dict, 'kde_plot')

        return bayesDict, ModelWeights_dict

    # -------------------------------------------------------------------------
    def generate_justData(self, modelDict):

        # Compute some variables
        metaModel = list(modelDict.items())[0][1]
        out_names = metaModel.ModelObj.Output.names
        data = np.array(
            [metaModel.ModelObj.observations[out] for out in out_names])

        # Generate data
        justData = np.tril(np.repeat(data, data.shape[1], axis=0))
        for key, metaModel in modelDict.items():
            data = np.array(
                [metaModel.ExpDesign.Y[out][0] for out in out_names])
            justData = np.vstack(
                (justData, np.tril(np.repeat(data, data.shape[1], axis=0)))
                )

        return justData

    # -------------------------------------------------------------------------
    def cal_modelWeight(self, BME_Dict):
        """
        Normalize the BME (Asumption: Model Prior weights are equal for models)

        Parameters
        ----------
        BME_Dict : TYPE
            DESCRIPTION.

        Returns
        -------
        all_ModelWeights : TYPE
            DESCRIPTION.
        ModelWeights : TYPE
            DESCRIPTION.

        """
        # Stack the BME values for all models
        all_BME = np.vstack(list(BME_Dict.values()))

        # Print statistics
        print('-'*10+"Statistical summary of BME"+'-'*10)
        print(np.quantile(np.log(all_BME), 0.25, axis=1) -
              np.quantile(np.log(all_BME), 0.5, axis=1))
        print(np.quantile(np.log(all_BME), 0.5, axis=1))
        print(np.quantile(np.log(all_BME), 0.75, axis=1) -
              np.quantile(np.log(all_BME), 0.5, axis=1))

        # Model weights
        modelWeights = np.divide(all_BME, np.nansum(all_BME, axis=0))

        print('\n'+'-'*10+"Statistical summary of Model Weights"+'-'*10)
        print(np.quantile(modelWeights, 0.25, axis=1) -
              np.quantile(modelWeights, 0.5, axis=1))
        print(np.quantile(modelWeights, 0.5, axis=1))
        print(np.quantile(modelWeights, 0.75, axis=1) -
              np.quantile(modelWeights, 0.5, axis=1))

        return modelWeights

    # -------------------------------------------------------------------------
    def plot_JustAnalysis(self, ModelWeights_dict):

        directory = 'Outputs_Comparison/'
        os.makedirs(directory, exist_ok=True)
        Color = [*mcolors.TABLEAU_COLORS]
        names = [*ModelWeights_dict]

        model_names = [model.replace('_', '$-$') for model in self.model_names]
        for name in names:
            fig, ax = plt.subplots()
            for i, model in enumerate(model_names[1:]):
                plt.plot(list(range(1, self.n_meas+1)),
                         ModelWeights_dict[name][i],
                         color=Color[i], marker='o',
                         ms=10, linewidth=2, label=model
                         )

            plt.title(name.replace('_', '$-$'))
            plt.ylabel("Weights")
            plt.xlabel("No. of measurement points")
            plt.legend(loc="best")
            fig.savefig(
                f'{directory}modelWeights_{name}.svg', bbox_inches='tight'
                )
            plt.close()

        # Confusion matrix for some measurement points
        for index in range(0, self.n_meas+1, 2):
            weights = np.array(
                [ModelWeights_dict[key][:, index] for key in ModelWeights_dict]
                )
            g = sns.heatmap(
                weights.T, annot=True, cmap='Blues', xticklabels=model_names,
                yticklabels=model_names[1:]
                )

            # x axis on top
            g.xaxis.tick_top()
            g.xaxis.set_label_position('top')
            g.set_xlabel(r"\textbf{Data generated by:}", labelpad=15)
            g.set_ylabel(r"\textbf{Model weight for:}", labelpad=15)
            g.figure.savefig(
                f"{directory}confusionMatrix_ND_{index+1}.svg",
                bbox_inches='tight'
                )
            plt.close()

    # -------------------------------------------------------------------------
    def boxPlot_ModelWeights(self, modelWeights, plotName):
        font_size = 40
        # mkdir for plots
        directory = 'Outputs_Comparison/'
        os.makedirs(directory, exist_ok=True)

        # Create figure
        fig, ax = plt.subplots()

        # Filter data using np.isnan
        mask = ~np.isnan(modelWeights.T)
        filtered_data = [d[m] for d, m in zip(modelWeights, mask.T)]

        # Create the boxplot
        bp = ax.boxplot(filtered_data, patch_artist=True, showfliers=False)

        # change outline color, fill color and linewidth of the boxes
        for box in bp['boxes']:
            # change outline color
            box.set(color='#7570b3', linewidth=4)
            # change fill color
            box.set(facecolor='#1b9e77')

        # change color and linewidth of the whiskers
        for whisker in bp['whiskers']:
            whisker.set(color='#7570b3', linewidth=2)

        # change color and linewidth of the caps
        for cap in bp['caps']:
            cap.set(color='#7570b3', linewidth=2)

        # change color and linewidth of the medians
        for median in bp['medians']:
            median.set(color='#b2df8a', linewidth=2)

        # change the style of fliers and their fill
        # for flier in bp['fliers']:
        #     flier.set(marker='o', color='#e7298a', alpha=0.75)

        # Custom x-axis labels
        model_names = [model.replace('_', '$-$') for model in self.model_names]
        ax.set_xticklabels(model_names)

        ax.set_ylabel('Weight', fontsize=font_size)

        # Title
        plt.title('Posterior Model Weights')

        # Set y lim
        ax.set_ylim((-0.05, 1.05))

        # Set size of the ticks
        for t in ax.get_xticklabels():
            t.set_fontsize(font_size)
        for t in ax.get_yticklabels():
            t.set_fontsize(font_size)

        # Save the figure
        fig.savefig(
            f'./{directory}ModelWeights{plotName}.pdf', bbox_inches='tight'
            )

        plt.close()

    # -------------------------------------------------------------------------
    def kdePlot_BayesFactor(self, BME_Dict, plotName):

        font_size = 40

        # mkdir for plots
        directory = 'Outputs_Comparison/'
        os.makedirs(directory, exist_ok=True)

        Colors = ["blue", "green", "gray", "brown"]

        model_names = list(BME_Dict.keys())
        nModels = len(model_names)

        # Plots
        fig, axes = plt.subplots(
            nrows=nModels, ncols=nModels, sharex=True, sharey=True
            )

        for i, key_i in enumerate(model_names):

            for j, key_j in enumerate(model_names):
                ax = axes[i, j]
                # Set size of the ticks
                for t in ax.get_xticklabels():
                    t.set_fontsize(font_size)
                for t in ax.get_yticklabels():
                    t.set_fontsize(font_size)

                if j != i:

                    # Null hypothesis: key_j is the better model
                    BayesFactor = np.log10(
                        np.divide(BME_Dict[key_i], BME_Dict[key_j])
                        )

                    # sns.kdeplot(BayesFactor, ax=ax, color=Colors[i], shade=True)
                    # sns.histplot(BayesFactor, ax=ax, stat="probability",
                    #              kde=True, element='step',
                    #              color=Colors[j])

                    # --https://stackoverflow.com/questions/55128462/how-to-normalize-seaborn-distplot ---
                    # taken from seaborn's source code (utils.py and distributions.py)
                    def seaborn_kde_support(data, bw, gridsize, cut, clip):
                        if clip is None:
                            clip = (-np.inf, np.inf)
                        support_min = max(data.min() - bw * cut, clip[0])
                        support_max = min(data.max() + bw * cut, clip[1])
                        return np.linspace(support_min, support_max, gridsize)

                    kde_estim = stats.gaussian_kde(
                        BayesFactor, bw_method='scott'
                        )

                    # manual linearization of data
                    # linearized = np.linspace(
                    #     quotient.min(), quotient.max(), num=500)

                    # or better: mimic seaborn's internal stuff
                    bw = kde_estim.scotts_factor() * np.std(BayesFactor)
                    linearized = seaborn_kde_support(
                        BayesFactor, bw, 100, 3, None)

                    # computes values of the estimated function on the
                    # estimated linearized inputs
                    Z = kde_estim.evaluate(linearized)

                    # https://stackoverflow.com/questions/29661574/normalize-
                    # numpy-array-columns-in-python
                    def normalize(x):
                        return (x - x.min(0)) / x.ptp(0)

                    # normalize so it is between 0;1
                    Z2 = normalize(Z)
                    ax.plot(linearized, Z2, "-", color=Colors[i], linewidth=4)
                    ax.fill_between(
                        linearized, 0, Z2, color=Colors[i], alpha=0.25
                        )

                    # Draw BF significant levels according to Jeffreys 1961
                    # Strong evidence for both models
                    ax.axvline(
                        x=np.log10(3), ymin=0, linewidth=4, color='dimgrey'
                        )
                    # Strong evidence for one model
                    ax.axvline(
                        x=np.log10(10), ymin=0, linewidth=4, color='orange'
                        )
                    # Decisive evidence for one model
                    ax.axvline(
                        x=np.log10(100), ymin=0, linewidth=4, color='r'
                        )

                    # legend
                    BF_label = key_i.replace('_', '$-$') + \
                        '/' + key_j.replace('_', '$-$')
                    legend_elements = [
                        patches.Patch(facecolor=Colors[i], edgecolor=Colors[i],
                                      label=f'BF({BF_label})')
                        ]
                    ax.legend(
                        loc='upper left', handles=legend_elements,
                        fontsize=font_size-(nModels+1)*5
                        )

                elif j == i:
                    # build a rectangle in axes coords
                    left, width = 0, 1
                    bottom, height = 0, 1

                    # axes coordinates are 0,0 is bottom left and 1,1 is upper
                    # right
                    p = patches.Rectangle(
                        (left, bottom), width, height, color='white',
                        fill=True, transform=ax.transAxes, clip_on=False
                        )
                    ax.grid(False)
                    ax.add_patch(p)
                    # ax.text(0.5*(left+right), 0.5*(bottom+top), key_i,
                    fsize = font_size+20 if nModels < 4 else font_size
                    ax.text(0.5, 0.5, key_i.replace('_', '$-$'),
                            horizontalalignment='center',
                            verticalalignment='center',
                            fontsize=fsize, color=Colors[i],
                            transform=ax.transAxes)

        # Defining custom 'ylim' values.
        custom_ylim = (0, 1.05)

        # Setting the values for all axes.
        plt.setp(axes, ylim=custom_ylim)

        # set labels
        for i in range(nModels):
            axes[-1, i].set_xlabel('log$_{10}$(BF)', fontsize=font_size)
            axes[i, 0].set_ylabel('Probability', fontsize=font_size)

        # Adjust subplots
        plt.subplots_adjust(wspace=0.2, hspace=0.1)

        plt.savefig(
            f'./{directory}BayesFactor{plotName}.svg', bbox_inches='tight'
            )
        plt.savefig(
            f'./{directory}BayesFactor{plotName}.pdf', bbox_inches='tight'
            )

        plt.close()
